from .cluster_test import (
    sequential_cluster_test_1samp,
    sequential_cluster_test_indep,
    sequential_cluster_test_corr
)

from .max_test import (
	sequential_permutation_t_test_1samp,
	sequential_permutation_test_indep,
	sequential_permutation_test_corr
)